/* -*- mode: C; c-file-style: "gnu"; indent-tabs-mode: nil; -*-
 *
 * Copyright (C) 2007 David Zeuthen <david@fubar.dk>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 *
 */

#ifdef HAVE_CONFIG_H
#  include "config.h"
#endif

#include <stdlib.h>
#include <stdarg.h>
#include <stdio.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/ioctl.h>
#include <fcntl.h>
#include <string.h>
#include <errno.h>

#include <linux/cdrom.h>

#include <libudev.h>

#include "poller.h"
#include "device.h"
#include "device-private.h"

#ifdef __linux__
extern char **environ;
static char **argv_buffer = NULL;
static size_t argv_size = 0;
#endif

/*#define POLL_SHOW_DEBUG*/

static void
set_proc_title_init (int argc, char *argv[])
{
#ifdef __linux__
  unsigned int i;
  char **new_environ, *endptr;

  /* This code is really really ugly. We make some memory layout
   * assumptions and reuse the environment array as memory to store
   * our process title in */

  for (i = 0; environ[i] != NULL; i++)
    ;

  endptr = i ? environ[i-1] + strlen (environ[i-1]) : argv[argc-1] + strlen (argv[argc-1]);

  argv_buffer = argv;
  argv_size = endptr - argv_buffer[0];

  /* Make a copy of environ */

  new_environ = malloc (sizeof(char*) * (i + 1));
  for (i = 0; environ[i] != NULL; i++)
    new_environ[i] = strdup (environ[i]);
  new_environ[i] = NULL;

  environ = new_environ;
#endif
}

/* this code borrowed from avahi-daemon's setproctitle.c (LGPL v2) */
static void
set_proc_title (const char *format,
                ...)
{
#ifdef __linux__
  size_t len;
  va_list ap;

  if (argv_buffer == NULL)
    goto out;

  va_start (ap, format);
  vsnprintf (argv_buffer[0], argv_size, format, ap);
  va_end (ap);

  len = strlen (argv_buffer[0]);

  memset (argv_buffer[0] + len, 0, argv_size - len);
  argv_buffer[1] = NULL;
 out:
  ;
#endif
}

/* ---------------------------------------------------------------------------------------------------- */

typedef struct
{
  int state; /* for ide-cd devices we need to keep track of the state */
  gboolean is_libata_cdrom;
  gboolean is_ide_cdrom;
  gchar *dev_path;
  gchar *uevent_path; /* only required for ide-cd */
} PollerDevice;

static GList *poller_devices_to_poll = NULL;

static guint poller_timeout_id = 0;

static void
poller_check_ide_cdrom (int fd, PollerDevice *device)
{
  int drive;
  gboolean trigger_change_uevent = FALSE;

  drive = ioctl (fd, CDROM_DRIVE_STATUS, CDSL_CURRENT);

  switch (drive)
    {
    case CDS_DISC_OK:
      if (device->state != 1)
        {
          trigger_change_uevent = TRUE;
          device->state = 1;
        }
      break;
    case CDS_NO_DISC:
      if (device->state != 2)
        {
          trigger_change_uevent = TRUE;
          device->state = 2;
        }
      break;
    case CDS_TRAY_OPEN:
      if (device->state != 3)
        {
          trigger_change_uevent = TRUE;
          device->state = 3;
        }
      break;
    case CDS_DRIVE_NOT_READY:
    case CDS_NO_INFO:
    case -1:
      if (device->state != 4)
        device->state = 4;
      break;
    default:
      if (device->state != 5)
        device->state = 5;
    }

  if (trigger_change_uevent)
    {
      int uevent_fd;
      uevent_fd = open (device->uevent_path, O_RDWR);
      if (uevent_fd != -1)
        {
#ifdef POLL_SHOW_DEBUG
          g_print ("**** POLLER (%d): triggering change uevent for %s\n", getpid (), device->dev_path);
#endif
          write (uevent_fd, "change", strlen("change"));
          close (uevent_fd);
        }
    }
}

static void
poller_poll_device (PollerDevice *device)
{
  int fd, fd2;

#ifdef POLL_SHOW_DEBUG
  g_print ("**** POLLER (%d): polling %s\n", getpid (), device->dev_path);
#endif

  if (device->is_libata_cdrom | device->is_ide_cdrom)
    {
      /* optical drives need special care
       *
       *  - use O_NONBLOCK to avoid closing the door
       *  - use O_EXCL to avoid interferring with cd burning software / audio playback / etc
       */
      fd = open (device->dev_path, O_RDONLY | O_NONBLOCK | O_EXCL);
      if (fd != -1)
        {
          if (device->is_ide_cdrom)
            {
#ifdef POLL_SHOW_DEBUG
          g_print ("**** POLLER (%d): open successful for %s\n", getpid (), device->dev_path);
#endif
          poller_check_ide_cdrom (fd, device);
            }
          close (fd);
        }
      else if (errno == EBUSY && device->is_ide_cdrom)
        {
          // TODO: check if media is mounted
          // in that case mounting with O_EXCL should be safe
#ifdef POLL_SHOW_DEBUG
          g_print ("**** POLLER (%d): open failed  %s\n", getpid (), strerror (errno) );
#endif
          fd = open (device->dev_path, O_RDONLY | O_NONBLOCK);
          if (fd != -1)
            {
#ifdef POLL_SHOW_DEBUG
              g_print ("**** POLLER (%d): open successful for %s without O_EXCL\n", getpid (), device->dev_path);
#endif
              poller_check_ide_cdrom (fd, device);
              close (fd);
            }
        }
    }
  else
    {
      fd = open (device->dev_path, O_RDONLY);
      fd2 = open (device->dev_path, O_RDONLY | O_NONBLOCK);
      if (fd != -1)
        {
          close (fd);
        }
      if (fd2 != -1)
        {
          close (fd2);
        }
    }
}

/* ---------------------------------------------------------------------------------------------------- */

static gboolean
poller_timeout_cb (gpointer user_data)
{
  g_list_foreach (poller_devices_to_poll, (GFunc)poller_poll_device, NULL);

  /* don't remove the source */
  return TRUE;
}

/* ---------------------------------------------------------------------------------------------------- */

static PollerDevice*
poller_device_populate (const gchar *dev_path)
{
  PollerDevice *device;

  device = g_malloc (sizeof(PollerDevice));
  device->dev_path = g_strdup (dev_path);
  device->uevent_path = NULL;
  device->state = 0;

  device->is_libata_cdrom = g_str_has_prefix (device->dev_path, "/dev/sr") || g_str_has_prefix (device->dev_path, "/dev/scd");

  if (device->is_libata_cdrom)
    {
      device->is_ide_cdrom = 0;
    }
  else
    { /* ask udev if device is a ide-cd cdrom */
      struct udev *udev;
      struct stat statbuf;
      struct udev_device *udevice;
      const gchar *type;
      const gchar *bus;

      udev = udev_new ();
      if (udev == NULL)
        {
          g_printerr ("Error initializing libudev: %m\n");
          goto out;
        }
      if (stat (device->dev_path, &statbuf) != 0)
        {
          g_printerr ("Error stating %s: %m\n", device->dev_path);
          goto out;
        }
      udevice = udev_device_new_from_devnum (udev, 'b', statbuf.st_rdev);
      if (udevice == NULL)
        {
          g_printerr ("Error getting udev device for %s: %m\n", device->dev_path);
          goto out;
        }
      type = udev_device_get_property_value (udevice, "ID_TYPE");
      bus = udev_device_get_property_value (udevice, "ID_BUS");

      device->is_ide_cdrom = (g_strcmp0 (type, "cd") == 0) && (g_strcmp0 (bus, "ata") == 0);
      if (device->is_ide_cdrom)
        {
          device->uevent_path = g_build_filename (udev_device_get_syspath (udevice), "uevent", NULL);
        }

    out:
      if (udevice != NULL)
        {
          udev_device_unref (udevice);
        }
      if (udev != NULL)
        {
          udev_unref (udev);
        }
    }

  return device;
}

static void
poller_device_free (PollerDevice *device)
{
  if (device == NULL)
    {
      return;
    }
  if (device->dev_path != NULL)
    {
#ifdef POLL_SHOW_DEBUG
      g_print ("**** POLLER (%d): freeing device '%s'\n", getpid (), device->dev_path);
#endif
      g_free (device->dev_path);
      device->dev_path = NULL;
    }
  if (device->uevent_path != NULL)
    {
      g_free (device->uevent_path);
      device->uevent_path = NULL;
    }
}

static gboolean
poller_have_data (GIOChannel *channel,
                  GIOCondition condition,
                  gpointer user_data)
{
  gchar *line;
  gsize line_length;
  GError *error;
  gint status;

  error = NULL;

  /* Exit if parent dies */
  if (condition == G_IO_HUP)
    {
      exit (1);
    }

 read_more:
  status = g_io_channel_read_line (channel, &line, &line_length, NULL, &error);
  if (error != NULL)
    {
      g_warning ("Error reading line from daemon: %s", error->message);
      g_error_free (error);
      goto out;
    }
  if (status == G_IO_STATUS_EOF || status == G_IO_STATUS_AGAIN)
    {
      goto out;
    }

  g_strstrip (line);

#ifdef POLL_SHOW_DEBUG
  g_print ("**** POLLER (%d): polling process read '%s'\n", getpid (), line);
#endif
  if (g_str_has_prefix (line, "set-poll:"))
    {
      gchar **tmp = NULL;
      int n;

      g_list_foreach (poller_devices_to_poll, (GFunc)poller_device_free, NULL);
      g_list_free (poller_devices_to_poll);
      poller_devices_to_poll = NULL;
      tmp = g_strsplit (line + strlen ("set-poll:"), " ", 0);
      for (n = 0; tmp != NULL && tmp[n] != NULL; n++)
        {
          poller_devices_to_poll = g_list_append (poller_devices_to_poll, poller_device_populate (tmp[n]));
        }
    }
  else
    {
      g_printerr ("**** POLLER (%d): unknown command '%s'\n", getpid (), line);
    }

  if (g_list_length (poller_devices_to_poll) == 0)
    {
      if (poller_timeout_id > 0)
        {
          g_source_remove (poller_timeout_id);
          poller_timeout_id = 0;
        }

      set_proc_title ("udisks-daemon: not polling any devices");
    }
  else
    {
      set_proc_title ("udisks-daemon: polling %s", line + strlen ("set-poll:"));

      if (poller_timeout_id == 0)
        {
          poller_timeout_id = g_timeout_add_seconds (2, poller_timeout_cb, NULL);
        }
    }

  g_free (line);
  goto read_more;

 out:
  /* keep the IOChannel around */
  return TRUE;
}

/* ---------------------------------------------------------------------------------------------------- */

static void
poller_run (gint fd)
{
  GMainLoop *loop;
  GIOChannel *io_channel;

  loop = g_main_loop_new (NULL, FALSE);

  io_channel = g_io_channel_unix_new (fd);
  g_io_channel_set_flags (io_channel, G_IO_FLAG_NONBLOCK, NULL);
  g_io_add_watch (io_channel, G_IO_IN | G_IO_HUP, poller_have_data, NULL);

  g_main_loop_run (loop);
}

/* ---------------------------------------------------------------------------------------------------- */

static gint poller_daemon_write_end_fd;

gboolean
poller_setup (int argc,
              char *argv[])
{
  gint pipefds[2];
  gboolean ret;

  ret = FALSE;

  if (pipe (pipefds) != 0)
    {
      g_warning ("Couldn't set up polling process, pipe() failed: %m");
      goto out;
    }

  switch (fork ())
    {
    case 0:
      /* child */
      close (pipefds[1]); /* close write end */
      set_proc_title_init (argc, argv);
      poller_run (pipefds[0]);
      break;

    default:
      /* parent */
      close (pipefds[0]); /* close read end */
      poller_daemon_write_end_fd = pipefds[1];
      break;

    case -1:
      g_warning ("Couldn't set up polling process, fork() failed: %m");
      goto out;
      break;
    }

  ret = TRUE;

 out:
  return ret;
}

/* ---------------------------------------------------------------------------------------------------- */

void
poller_set_devices (GList *devices)
{
  GList *l;
  gchar **device_array;
  guint n;
  gchar *joined;
  gchar *devices_to_poll;
  static gchar *devices_currently_polled = NULL;

  device_array = g_new0 (gchar *, g_list_length (devices) + 2);

  for (l = devices, n = 0; l != NULL; l = l->next)
    {
      Device *device = DEVICE (l->data);

      device_array[n++] = device->priv->device_file;
    }

  g_qsort_with_data (device_array, n, sizeof(gchar *), (GCompareDataFunc) g_strcmp0, NULL);

  device_array[n] = "\n";

  joined = g_strjoinv (" ", device_array);
  g_free (device_array);
  devices_to_poll = g_strconcat ("set-poll:", joined, NULL);
  g_free (joined);

  /* only poke the polling process if the list of currently polled devices change */
  if (g_strcmp0 (devices_to_poll, devices_currently_polled) != 0)
    {
      g_free (devices_currently_polled);
      devices_currently_polled = devices_to_poll;

#ifdef POLL_SHOW_DEBUG
      g_print ("**** POLLER (%d): Sending poll command: '%s'\n", getpid (), devices_currently_polled);
#endif
      if (write (poller_daemon_write_end_fd, devices_currently_polled, strlen (devices_currently_polled)) < 0)
        g_error ("**** POLLER (%d): Failed to send polled devices: %s", getpid (), g_strerror (errno));
    }
  else
    {
      g_free (devices_to_poll);
    }
}

/* ---------------------------------------------------------------------------------------------------- */
